#include <stdint.h>
#include "fip_urv_regs.h"
#include "fip_urv_cmd.h"
#include "hydra_supervisor_regs.h"

#define SUPERVISOR_BASE 0x80000
#define SUPERVISOR ((volatile struct hydra_supervisor_regs *)SUPERVISOR_BASE)
#define WD_KEY 0xc0423bc9

#define REGS ((volatile struct fip_urv_regs *)0x100000)

/* I2C open-core registers.  */
#define OC_I2C_PRES_LO 0
#define OC_I2C_PRES_HI 1
#define OC_I2C_CTRL    2
#define OC_I2C_RXTX    3
#define OC_I2C_STAT    4
#define OC_I2C_CR      4
#define OC_I2C_CFG     5

/* Produced variable length (the reply).  */
#define VAR_LENGTH 16

/* Save regs to memory to detect any divergence.  */
extern void refresh_regs(void);

static unsigned extract_word (volatile uint32_t *buf)
{
  unsigned res;

  /* Little endian.  */
  res = (buf[0] & 0xff) << 0;
  res |= (buf[1] & 0xff) << 8;
  res |= (buf[2] & 0xff) << 16;
  res |= (buf[3] & 0xff) << 24;

  return res;
}

static void insert_word(unsigned char *buf, unsigned val)
{
  buf[0] = (val >> 0) & 0xff;
  buf[1] = (val >> 8) & 0xff;
  buf[2] = (val >> 16) & 0xff;
  buf[3] = (val >> 24) & 0xff;
}

struct i2c_cmd {
  unsigned char state;
  unsigned char len;
  unsigned char coff;
  unsigned char boff;
  unsigned char buf[32];
  const unsigned char *cmds;
} __attribute__((aligned(4)));

#define MONIMOD_I2C_ADDR (0x19 << 1)

/* i2c commands that are executed to get values from monimod.  */
static const unsigned char monimod_cmds[] =
  {
   I2C_CMD_START, MONIMOD_I2C_ADDR,
   I2C_CMD_WRITE_STOP | (2 << 4), 0x00, 0x00,  /* page 0 */

   I2C_CMD_START, MONIMOD_I2C_ADDR,
   I2C_CMD_WRITE | (1 << 4), 0x8b, /* vout */
   I2C_CMD_START, MONIMOD_I2C_ADDR | 1,
   I2C_CMD_READ_STOP | (2 << 4),

   I2C_CMD_START, MONIMOD_I2C_ADDR,
   I2C_CMD_WRITE | (1 << 4), 0x8c, /* iout */
   I2C_CMD_START, MONIMOD_I2C_ADDR | 1,
   I2C_CMD_READ_STOP | (2 << 4),

   I2C_CMD_START, MONIMOD_I2C_ADDR,
   I2C_CMD_WRITE_STOP | (2 << 4), 0x00, 0x01,  /* page 1 */

   I2C_CMD_START, MONIMOD_I2C_ADDR,
   I2C_CMD_WRITE | (1 << 4), 0x8b, /* vout */
   I2C_CMD_START, MONIMOD_I2C_ADDR | 1,
   I2C_CMD_READ_STOP | (2 << 4),

   I2C_CMD_START, MONIMOD_I2C_ADDR,
   I2C_CMD_WRITE_STOP | (2 << 4), 0x00, 0x02,  /* page 2 */

   I2C_CMD_START, MONIMOD_I2C_ADDR,
   I2C_CMD_WRITE | (1 << 4), 0x8b, /* vout */
   I2C_CMD_START, MONIMOD_I2C_ADDR | 1,
   I2C_CMD_READ_STOP | (2 << 4),

   I2C_CMD_START, MONIMOD_I2C_ADDR,
   I2C_CMD_WRITE_STOP | (2 << 4), 0x00, 0x03,  /* page 3 */

   I2C_CMD_START, MONIMOD_I2C_ADDR,
   I2C_CMD_WRITE | (1 << 4), 0x8b, /* vout */
   I2C_CMD_START, MONIMOD_I2C_ADDR | 1,
   I2C_CMD_READ_STOP | (2 << 4),

   I2C_CMD_START, MONIMOD_I2C_ADDR,
   I2C_CMD_WRITE | (1 << 4), 0x8d, /* temp-1 */
   I2C_CMD_START, MONIMOD_I2C_ADDR | 1,
   I2C_CMD_READ_STOP | (2 << 4),

   I2C_CMD_START, MONIMOD_I2C_ADDR,
   I2C_CMD_WRITE | (1 << 4), 0x8e, /* temp-2 */
   I2C_CMD_START, MONIMOD_I2C_ADDR | 1,
   I2C_CMD_READ_STOP | (2 << 4),

   I2C_CMD_START, MONIMOD_I2C_ADDR,
   I2C_CMD_WRITE | (1 << 4), 0x8f, /* temp-3 */
   I2C_CMD_START, MONIMOD_I2C_ADDR | 1,
   I2C_CMD_READ_STOP | (2 << 4),

   I2C_CMD_IDLE
  };

static unsigned char monimod_buf[16];
static unsigned char monimod_cycle;

/* Use of i2c for monimod scan.
   0: never (allowed for user)
   1: scan (forbidden to user)
   2: scanning.
   3: in reset. */
static unsigned char i2c_monimod;


struct status_t {
  /* Number of reset to monimod.  */
  unsigned monimod_reset;
  /* Minimum watchdog count.  */
  unsigned min_wd_count;
};

static unsigned stat_regs[STAT_NBR];

static inline void
do_i2c_write (struct i2c_cmd *i2c)
{
  /* Set the stop bit only on the last byte and if requested.  */
  unsigned char stop =
    i2c->len == 1 && i2c->state == I2C_CMD_WRITE_STOP;
  REGS->i2c[OC_I2C_RXTX] = i2c->cmds[i2c->coff];
  REGS->i2c[OC_I2C_CR] = 0x10 | (stop ? 0x40 : 0);  /* WRITE */
  i2c->coff++;
}

static inline void
do_i2c_read (struct i2c_cmd *i2c)
{
  /* Set the stop bit only on the last byte and if requested.  */
  unsigned char stop =
    i2c->len == 1 && i2c->state == I2C_CMD_READ_STOP;
  REGS->i2c[OC_I2C_CR] = 0x20 | (stop ? 0x48 : 0);  /* READ */
}

static void
do_i2c_done (struct i2c_cmd *i2c, unsigned char stat)
{
  /* Save values in case of error.  */
  i2c->buf[31] = stat;
  i2c->buf[30] = i2c->state;

  /* According to the current command (state) check if it has been
     executed, and the status.  */
  if (i2c->state == I2C_CMD_START) {
    if (stat == 0x41) {
      /* Busy + Interrupt, ok.  */
      i2c->state = I2C_CMD_RUN;
      return;
    }
  }
  else if (i2c->state == I2C_CMD_WRITE
           || i2c->state == I2C_CMD_WRITE_STOP) {
    if (stat == 0x41 || (stat == 0x01 && i2c->len == 1)) {
      /* Next byte.  */
      i2c->len--;
      if (i2c->len == 0)
        i2c->state = I2C_CMD_RUN;
      else
        do_i2c_write(i2c);
      return;
    }
  }
  else if (i2c->state == I2C_CMD_READ
           || i2c->state == I2C_CMD_READ_STOP) {
    /* We are the master.  No need to check STAT, as the ACK/BUSY
       bits are in fact set by us.  */
    /* Read the byte received.  */
    i2c->buf[i2c->boff] = REGS->i2c[OC_I2C_RXTX];
    i2c->boff++;
    i2c->len--;
    if (i2c->len == 0) {
      /* All the bytes have been read, next i2c command.  */
      i2c->state = I2C_CMD_RUN;
    }
    else {
      /* Prepare to read next byte.  */
      do_i2c_read (i2c);
    }
    return;
  }
  else if (i2c->state == I2C_CMD_READLEN) {
    if (stat == 0x41) {
      i2c->len = REGS->i2c[OC_I2C_RXTX];
      i2c->buf[i2c->boff] = i2c->len;
      i2c->boff++;
      i2c->state = I2C_CMD_RUN;
      return;
    }
  }

  i2c->state = I2C_CMD_ERR;
  REGS->i2c[OC_I2C_CR] = 0x40; /* Stop */

  if (i2c_monimod == 2) {
    /* This is an error during monimod scan, shutdown monimod */
    REGS->monimod_shdn = 1;
    stat_regs[STAT_MONIMOD_RST]++;
    i2c_monimod = 3;
  }
}

static void
do_i2c_run (struct i2c_cmd *i2c)
{
  /* Read the command to be executed.  */
  unsigned char cmd = i2c->cmds[i2c->coff];

  /* Default state: the command.  */
  i2c->state = cmd & 0x0f;

  if (cmd & 0xf0)
    i2c->len = cmd >> 4;

  if ((cmd & 0x0f) == I2C_CMD_IDLE) {
    /* Last command.  */
    if (i2c_monimod == 2) {
      /* Was for monitor scanning.  */
      unsigned i;
      /* Copy result, so that a consistent view can be extracted.  */
      for (i = 0; i < sizeof(monimod_buf); i++)
        monimod_buf[i] = i2c->buf[i];
      /* Leave that mode. */
      i2c_monimod = 1;
      /* Change cycle.  */
      monimod_cycle++;
    }
  }
  else if (cmd == I2C_CMD_START) {
    REGS->i2c[OC_I2C_RXTX] = i2c->cmds[i2c->coff + 1];
    REGS->i2c[OC_I2C_CR] = 0x90;  /* START + WRITE */
    i2c->coff += 2;
  }
  else if ((cmd & 0x0f) == I2C_CMD_WRITE
           || (cmd & 0x0f) == I2C_CMD_WRITE_STOP) {
    i2c->coff++;
    do_i2c_write(i2c);
  }
  else if ((cmd & 0x0f) == I2C_CMD_READ
           || (cmd & 0x0f) == I2C_CMD_READ_STOP
           || (cmd & 0x0f) == I2C_CMD_READLEN) {
    i2c->coff++;
    do_i2c_read(i2c);
  }
  else {
    /* Bad command.  */
    i2c->state = I2C_CMD_ERR;
  }
}

int
main (void)
{
  unsigned timeout_led;
  unsigned char reply[VAR_LENGTH];
  unsigned char rdy;
  unsigned char leds;
  unsigned char rlen;
  static struct i2c_cmd i2c;
  static unsigned timeout_mon;
  static unsigned char user_i2c_cmds[16];

#if VAR_LENGTH == 8
  /* That's the default.  */
#elif VAR_LENGTH == 16
  REGS->fip_var = 2;
  SUPERVISOR->wd_period = 5000;
#elif VAR_LENGTH == 64
  REGS->fip_var = 4;
  SUPERVISOR->wd_period = 5000;
#elif VAR_LENGTH == 124
  REGS->fip_var = 5;
  SUPERVISOR->wd_period = 8500;
#else
# error "Incorrect VAR_LENGTH"
#endif

  timeout_led = 0;
  rdy = 0;

  /* Var1 has been consumed so it's released in case of timeout. */
  REGS->fip_var1 = 0;

  /* Clear recovery bit.  */
  SUPERVISOR->cpu = 0;

  if (SUPERVISOR->reset_cause == 0)
    {
      /* Init.  */
      leds = 1;
      i2c.state = I2C_CMD_IDLE;
      i2c.coff = 0;
      i2c.boff = 0;
      i2c.len = 0;
      i2c.cmds = user_i2c_cmds;

      timeout_mon = 0;
      monimod_cycle = 0;
      i2c_monimod = 1;
      stat_regs[STAT_MIN_WD_CNT] = ~0;

      /* Power-on monimod */
      REGS->monimod_shdn = 0;

      /* Enable scrubber */
      SUPERVISOR->scrub_cfg = HYDRA_SUPERVISOR_REGS_SCRUB_CFG_IRAM_EN
        | HYDRA_SUPERVISOR_REGS_SCRUB_CFG_DRAM_EN;

      /* Init i2c: prescaler (400khz) */
      REGS->i2c[OC_I2C_PRES_LO] = 25;
      REGS->i2c[OC_I2C_PRES_HI] = 0;
      REGS->i2c[OC_I2C_CTRL] = 0x80;  /* Enable */
    }
  else {
    leds = 3;
    stat_regs[STAT_RESET]++;
  }

  /* Infinite loop.  */
  while (1)
    {
      unsigned cpu;

      /* Build date stat */
      stat_regs[STAT_DATE_ADDR] = C_BUILD_DATE;
      stat_regs[STAT_HASH_ADDR] = C_GIT_HASH;

      /* Check for FIP message.  */
      unsigned int fip_status;

      /* Get minimum watchdog count */
      unsigned count = SUPERVISOR->wd_count;
      if (count < stat_regs[STAT_MIN_WD_CNT])
        stat_regs[STAT_MIN_WD_CNT] = count;

      SUPERVISOR->wd_key = WD_KEY;

      refresh_regs();

      stat_regs[STAT_CYCLE_CNT]++;

      /* Force recovery in case of divergence */
      cpu = SUPERVISOR->cpu;
      if (cpu != 0) {
        stat_regs[STAT_RECOVERY]++;
        stat_regs[STAT_LAST_RST] = cpu;
        asm volatile ("");
        SUPERVISOR->cpu = HYDRA_SUPERVISOR_REGS_CPU_RECOVERY;
      }

      fip_status = REGS->fip_status;

      /* Control leds.  */
      if (leds & 0x80)
        REGS->leds = leds;
      else
        {
          /* Leds 3 & 2 are for liveness */
          REGS->leds = leds;
          timeout_led++;
          if (timeout_led == 100000)
            {
              leds = (leds & 1) ^ 1;
              timeout_led = 0;
            }
        }

      if (fip_status & 1)
        {
          /* Variable received (from FIP).  */
          unsigned char cmd;
          unsigned char id;

          /* Consume variable.  */
          REGS->fip_var1 = 1;
          cmd = REGS->fip_reg[0x02];
          id = REGS->fip_reg[0x03];

          /* Default reply: command + id.  */
          reply[0] = cmd;
          reply[1] = id;
          reply[3] = REGS->fip_reg[0x02 + VAR_LENGTH]; /* MPS */
          /* MPS: bit 0: refreshment, bit 2: significance */

          /* Default reply length.  */
          rlen = 8;

          stat_regs[STAT_FIP_CMD]++;

          /* No switch (smaller)  */
          if (cmd == CMD_SET_LEDS) {
              leds = REGS->fip_reg[0x06];
          }
          else if (cmd == CMD_READ_WORD) {
            volatile unsigned *addr;

            addr = (unsigned *)extract_word(&REGS->fip_reg[0x6]);
            insert_word(&reply[4], *addr);
          }
          else if (cmd == CMD_WRITE_WORD) {
            volatile unsigned *addr;
            unsigned val;
            unsigned k;
            unsigned l;

            stat_regs[STAT_FIP_WRITES]++;

            l = REGS->fip_reg[0x02 + 2];
            addr = (unsigned *)extract_word(&REGS->fip_reg[0x02 + 4]);

            stat_regs[STAT_WRITE_ADDR] = (unsigned)addr;

            for (k = 4; k < l; k += 4)
              {
                val = extract_word(&REGS->fip_reg[0x02 + 0x04 + k]);
                addr[(k - 4) >> 2] = val;
              }
          }
          else if (cmd == CMD_WRITE_ECC) {
            volatile unsigned *addr;
            unsigned val;
            unsigned ecc;

            stat_regs[STAT_FIP_WRITES]++;

            addr = (unsigned *)extract_word(&REGS->fip_reg[0x6]);

            stat_regs[STAT_WRITE_ADDR] = (unsigned)addr | 1;

            val = extract_word(&REGS->fip_reg[0x6 + 4]);
            ecc = REGS->fip_reg[0x6 + 8];
            SUPERVISOR->dram_ecc_mask = ecc;
            *addr = val;
            SUPERVISOR->dram_ecc_mask = 0;
          }
          else if (cmd == CMD_I2C_MON) {
            unsigned val = REGS->fip_reg[0x06];

            if ((i2c_monimod == 0 && val == 1)
                || (i2c_monimod == 1 && val == 0))
              i2c_monimod = val;
            insert_word(&reply[4], i2c_monimod);
          }
          else if (cmd == CMD_I2C_TFR) {
            if (i2c_monimod != 0 || (i2c.state & 0x0f) != I2C_CMD_IDLE) {
              /* Busy */
              reply[4] = 1;
            }
            else {
              /* Extract i2c ommands from fip.  */
              unsigned k;
              unsigned clen = REGS->fip_reg[0x6];
              for (k = 0; k < clen; k++)
                user_i2c_cmds[k] = REGS->fip_reg[0x7 + k];
              user_i2c_cmds[clen] = I2C_CMD_IDLE;
              i2c.coff = 0;
              i2c.boff = 0;
              i2c.cmds = user_i2c_cmds;
              /* Has to execute i2c commands.  */
              i2c.state = I2C_CMD_RUN;
              /* fip reply.  */
              reply[4] = 0;
            }
          }
          else if (cmd == CMD_I2C_STAT) {
            reply[4] = i2c.state;
            reply[5] = i2c.len;
            reply[6] = i2c.coff;
            reply[7] = i2c.boff;
          }
          else if (cmd == CMD_I2C_BUF) {
            unsigned off = REGS->fip_reg[0x6];
            reply[4] = i2c.buf[off + 0];
            reply[5] = i2c.buf[off + 1];
            reply[6] = i2c.buf[off + 2];
            reply[7] = i2c.buf[off + 3];
          }
          else if (cmd == CMD_MON_BUF) {
            unsigned off = REGS->fip_reg[0x6];
            off <<= 1;
            reply[4] = monimod_buf[off + 0];
            reply[5] = monimod_buf[off + 1];
            reply[6] = 0;
            reply[7] = monimod_cycle;
          }
          else if (cmd == CMD_STAT_BUF) {
            unsigned off = REGS->fip_reg[0x6];
            insert_word(&reply[4], stat_regs[off]);
          }
          else if (cmd == CMD_CPU_FLT) {
            unsigned cpu = REGS->fip_reg[0x6];
            if (SUPERVISOR->force_divergence != cpu)
              reply[4] = 1;
          }
          else if (cmd == CMD_FIP_PING) {
            unsigned i;
            rlen = VAR_LENGTH;
            reply[2] = rlen;
            for (i = 4; i < rlen; i++)
              reply[i] = ~REGS->fip_reg[0x02 + i];
          }
          else if (cmd == CMD_WAIT) {
            unsigned val;
            unsigned param;

            val = extract_word(&REGS->fip_reg[0x6]);
            param = extract_word(&REGS->fip_reg[0x6 + 4]);
            SUPERVISOR->wd_key = WD_KEY;

            while (val-- != 0)
              if (param & 1)
                SUPERVISOR->wd_key = WD_KEY;
          }
          else {
            reply[0] = CMD_ERROR;
            stat_regs[STAT_FIP_PAD]++;
          }

          /* Ready to send a reply.  */
          rdy = 1;
          /* Variable has been consumed.  */
          REGS->fip_var1 = 0;
        }

      if ((fip_status & 2) && rdy)
        {
          int i;

          /* Produce variable.  */
          REGS->fip_var3 = 1;

          for (i = 0; i < rlen; i++)
            REGS->fip_reg[0x102 + i] = reply[i];

          REGS->fip_var3 = 0;

          rdy = 0;
        }

      if ((i2c.state & 0x0f) != I2C_CMD_IDLE) {
        if (i2c.state != I2C_CMD_RUN) {
          /* For end of transfer, check TIP (transfer in progess)  */
          unsigned char stat = REGS->i2c[OC_I2C_STAT];
          if (!(stat & 2))
            do_i2c_done (&i2c, stat);
        }

        /* Execute the command.
           As do_i2c_done can change the state, the state needs to be checked
           again.  */
        if (i2c.state == I2C_CMD_RUN)
          do_i2c_run (&i2c);
      }
      else if (i2c_monimod == 1 && timeout_mon++ == 200000) {
        timeout_mon = 0;

        /* Select i2c 0 (for monimod).  */
        REGS->i2c[OC_I2C_CFG] = 0;

        /* Start monimod commands.  */
        i2c.cmds = monimod_cmds;
        i2c.coff = 0;
        i2c.boff = 0;
        i2c.state = I2C_CMD_RUN;

        /* Scanning.  */
        i2c_monimod = 2;
      }
      else if (i2c_monimod == 3 && timeout_mon++ == 100000) {
        /* Was in reset (power-off), re-enable monimod.  */
        REGS->monimod_shdn = 0;
        i2c_monimod = 1;
        timeout_mon = 0;
      }
    }
}
